import unittest
from textwrap import dedent

from click import testing as click_testing


def get_runtests():
    # Lazy load to avoid circular imports.

    from .._cli.main import pysr

    class TestCli(unittest.TestCase):
        # TODO: Include test for custom project here.
        def setUp(self):
            self.cli_runner = click_testing.CliRunner()

        def test_help_on_all_commands(self):
            expected = dedent(
                """
                    Usage: pysr [OPTIONS] COMMAND [ARGS]...

                    Options:
                      --help  Show this message and exit.

                    Commands:
                      install  DEPRECATED (dependencies are now installed at import).
                      test     Run parts of the PySR test suite.
                """
            )
            result = self.cli_runner.invoke(pysr, ["--help"])
            self.assertEqual(result.output.strip(), expected.strip())
            self.assertEqual(result.exit_code, 0)

        def test_help_on_install(self):
            expected = dedent(
                """
                Usage: pysr install [OPTIONS]

                  DEPRECATED (dependencies are now installed at import).

                Options:
                  -p, --project TEXT
                  -q, --quiet         Disable logging.
                  --precompile
                  --no-precompile
                  --help              Show this message and exit.
                """
            )
            result = self.cli_runner.invoke(pysr, ["install", "--help"])
            self.assertEqual(result.output.strip(), expected.strip())
            self.assertEqual(result.exit_code, 0)

        def test_help_on_test(self):
            expected = dedent(
                """
                Usage: pysr test [OPTIONS] TESTS

                  Run parts of the PySR test suite.

                  Choose from main, jax, torch, cli, dev, and startup. You can give multiple
                  tests, separated by commas.

                Options:
                  -k TEXT  Filter expressions to select specific tests.
                  --help   Show this message and exit.
                """
            )
            result = self.cli_runner.invoke(pysr, ["test", "--help"])
            self.assertEqual(result.output.strip(), expected.strip())
            self.assertEqual(result.exit_code, 0)

    def runtests(just_tests=False):
        """Run all tests in cliTest.py."""
        tests = [TestCli]
        if just_tests:
            return tests
        loader = unittest.TestLoader()
        suite = unittest.TestSuite()
        for test in tests:
            suite.addTests(loader.loadTestsFromTestCase(test))
        runner = unittest.TextTestRunner()
        return runner.run(suite)

    return runtests



"""Functions to help export PySR equations to LaTeX."""

import pandas as pd
import sympy  # type: ignore
from sympy.printing.latex import LatexPrinter  # type: ignore


class PreciseLatexPrinter(LatexPrinter):
    """Modified SymPy printer with custom float precision."""

    def __init__(self, settings=None, prec=3):
        super().__init__(settings)
        self.prec = prec

    def _print_Float(self, expr):
        # Reduce precision of float:
        reduced_float = sympy.Float(expr, self.prec)
        return super()._print_Float(reduced_float)


def sympy2latex(expr, prec=3, full_prec=True, **settings) -> str:
    """Convert sympy expression to LaTeX with custom precision."""
    settings["full_prec"] = full_prec
    printer = PreciseLatexPrinter(settings=settings, prec=prec)
    return str(printer.doprint(expr))


def generate_table_environment(
    columns: list[str] = ["equation", "complexity", "loss"]
) -> tuple[str, str]:
    margins = "c" * len(columns)
    column_map = {
        "complexity": "Complexity",
        "loss": "Loss",
        "equation": "Equation",
        "score": "Score",
    }
    columns = [column_map[col] for col in columns]
    top_pieces = [
        r"\begin{table}[h]",
        r"\begin{center}",
        r"\begin{tabular}{@{}" + margins + r"@{}}",
        r"\toprule",
        " & ".join(columns) + r" \\",
        r"\midrule",
    ]

    bottom_pieces = [
        r"\bottomrule",
        r"\end{tabular}",
        r"\end{center}",
        r"\end{table}",
    ]
    top_latex_table = "\n".join(top_pieces)
    bottom_latex_table = "\n".join(bottom_pieces)

    return top_latex_table, bottom_latex_table


def sympy2latextable(
    equations: pd.DataFrame,
    indices: list[int] | None = None,
    precision: int = 3,
    columns: list[str] = ["equation", "complexity", "loss", "score"],
    max_equation_length: int = 50,
    output_variable_name: str = "y",
) -> str:
    """Generate a booktabs-style LaTeX table for a single set of equations."""
    assert isinstance(equations, pd.DataFrame)

    latex_top, latex_bottom = generate_table_environment(columns)
    latex_table_content = []

    if indices is None:
        indices = list(equations.index)

    for i in indices:
        latex_equation = sympy2latex(
            equations.iloc[i]["sympy_format"],
            prec=precision,
        )
        complexity = str(equations.iloc[i]["complexity"])
        loss = sympy2latex(
            sympy.Float(equations.iloc[i]["loss"]),
            prec=precision,
        )
        score = sympy2latex(
            sympy.Float(equations.iloc[i]["score"]),
            prec=precision,
        )

        row_pieces = []
        for col in columns:
            if col == "equation":
                if len(latex_equation) < max_equation_length:
                    row_pieces.append(
                        "$" + output_variable_name + " = " + latex_equation + "$"
                    )
                else:
                    broken_latex_equation = " ".join(
                        [
                            r"\begin{minipage}{0.8\linewidth}",
                            r"\vspace{-1em}",
                            r"\begin{dmath*}",
                            output_variable_name + " = " + latex_equation,
                            r"\end{dmath*}",
                            r"\end{minipage}",
                        ]
                    )
                    row_pieces.append(broken_latex_equation)

            elif col == "complexity":
                row_pieces.append("$" + complexity + "$")
            elif col == "loss":
                row_pieces.append("$" + loss + "$")
            elif col == "score":
                row_pieces.append("$" + score + "$")
            else:
                raise ValueError(f"Unknown column: {col}")

        latex_table_content.append(
            " & ".join(row_pieces) + r" \\",
        )

    return "\n".join([latex_top, *latex_table_content, latex_bottom])


def sympy2multilatextable(
    equations: list[pd.DataFrame],
    indices: list[list[int]] | None = None,
    precision: int = 3,
    columns: list[str] = ["equation", "complexity", "loss", "score"],
    output_variable_names: list[str] | None = None,
) -> str:
    """Generate multiple latex tables for a list of equation sets."""
    # TODO: Let user specify custom output variable

    latex_tables = [
        sympy2latextable(
            equations[i],
            (None if not indices else indices[i]),
            precision=precision,
            columns=columns,
            output_variable_name=(
                "y_{" + str(i) + "}"
                if output_variable_names is None
                else output_variable_names[i]
            ),
        )
        for i in range(len(equations))
    ]

    return "\n\n".join(latex_tables)


def with_preamble(table_string: str) -> str:
    preamble_string = [
        r"\usepackage{breqn}",
        r"\usepackage{booktabs}",
        "",
        "...",
        "",
        table_string,
    ]
    return "\n".join(preamble_string)



# Load YAML file param_groupings.yml:
import re
import sys

from docstring_parser import parse
from yaml import safe_load

sys.path.append("..")


from pysr import PySRRegressor

found_params = []


def str_param_groups(param_groupings, params, cur_heading=2):
    global found_params
    # Recursively print the parameter descriptions, defaults,
    # with headings from the param groupings dict.
    if isinstance(param_groupings, list):
        return "\n\n".join(
            str_param_groups(param, params, cur_heading) for param in param_groupings
        )
    elif isinstance(param_groupings, dict):
        for heading, param_grouping in param_groupings.items():
            return (
                f"{'#' * cur_heading} {heading}"
                + "\n\n"
                + str_param_groups(param_grouping, params, cur_heading + 1)
            )
    elif isinstance(param_groupings, str):
        found_params.append(param_groupings)

        default_value = re.search(
            r"Default is `(.*)`", params[param_groupings].description
        )
        clean_desc = re.sub(r"Default is .*", "", params[param_groupings].description)
        # Prepend every line with 4 spaces:
        clean_desc = "\n".join("    " + line for line in clean_desc.splitlines())
        return (
            f"  - **`{param_groupings}`**"
            + "\n\n"
            + clean_desc
            + (
                "\n\n    " + f"*Default:* `{default_value.group(1)}`"
                if default_value
                else ""
            )
        )
    else:
        raise TypeError(f"Unexpected type {type(param_groupings)}")


if __name__ == "__main__":
    # This is the path to the param_groupings.yml file
    # relative to the current file.
    path = "../pysr/param_groupings.yml"
    with open(path, "r") as f:
        param_groupings = safe_load(f)

    # This is basically a dict of lists and dicts.

    # Let's load in the parameter descriptions from the docstring of PySRRegressor:
    raw_params = parse(PySRRegressor.__doc__).params
    params = {
        param.arg_name: param
        for param in raw_params
        if param.arg_name[-1] != "_" and param.arg_name != "**kwargs"
    }

    output = str_param_groups(param_groupings, params, cur_heading=3)
    assert len(set(found_params) ^ set(params.keys())) == 0
    print("## PySRRegressor Parameters")
    print(output)



"""Functions for initializing the Julia environment and installing deps."""

from typing import Any, Callable, cast, overload

import numpy as np
from juliacall import convert as jl_convert  # type: ignore
from numpy.typing import NDArray

from .deprecated import init_julia, install
from .julia_import import AnyValue, jl

jl_convert = cast(Callable[[Any, Any], Any], jl_convert)

jl.seval("using Serialization: Serialization")
jl.seval("using PythonCall: PythonCall")

Serialization = jl.Serialization
PythonCall = jl.PythonCall

jl.seval("using SymbolicRegression: plus, sub, mult, div, pow")


def _escape_filename(filename):
    """Turn a path into a string with correctly escaped backslashes."""
    if filename is None:
        return None
    str_repr = str(filename)
    str_repr = str_repr.replace("\\", "\\\\")
    return str_repr


def _load_cluster_manager(cluster_manager: str):
    jl.seval(f"using ClusterManagers: addprocs_{cluster_manager}")
    return jl.seval(f"addprocs_{cluster_manager}")


def jl_array(x, dtype=None):
    if x is None:
        return None
    elif dtype is None:
        return jl_convert(jl.Array, x)
    else:
        return jl_convert(jl.Array[dtype], x)


def jl_dict(x):
    return jl_convert(jl.Dict, x)


def jl_is_function(f) -> bool:
    return cast(bool, jl.seval("op -> op isa Function")(f))


def jl_serialize(obj: Any) -> NDArray[np.uint8]:
    buf = jl.IOBuffer()
    Serialization.serialize(buf, obj)
    return np.array(jl.take_b(buf))


@overload
def jl_deserialize(s: NDArray[np.uint8]) -> AnyValue: ...
@overload
def jl_deserialize(s: None) -> None: ...
def jl_deserialize(s):
    if s is None:
        return s
    buf = jl.IOBuffer()
    jl.write(buf, jl_array(s))
    jl.seekstart(buf)
    return Serialization.deserialize(buf)



import copy
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, NewType, TypeAlias

import numpy as np
import pandas as pd

from .export import add_export_formats
from .julia_helpers import jl_array
from .julia_import import AnyValue, SymbolicRegression, jl

# For type checking purposes
if TYPE_CHECKING:
    from .sr import PySRRegressor  # pragma: no cover

    PySRRegressor: TypeAlias = PySRRegressor  # pragma: no cover
else:
    PySRRegressor = NewType("PySRRegressor", Any)


class AbstractExpressionSpec(ABC):
    """Abstract base class describing expression types.

    This basically just holds the options for the expression type,
    as well as explains how to parse and evaluate them.

    All expression types must implement:

    1. julia_expression_type(): The actual expression type, returned as a Julia object.
        This will get stored as `expression_type` in `SymbolicRegression.Options`.
    2. julia_expression_options(): Method to create the expression options, returned as a Julia object.
        These will get stored as `expression_options` in `SymbolicRegression.Options`.
    3. create_exports(), which will be used to create the exports of the equations, such as
        the executable format, the SymPy format, etc.

    It may also optionally implement:

    - supports_sympy, supports_torch, supports_jax, supports_latex: Whether this expression type supports the corresponding export format.
    """

    @abstractmethod
    def julia_expression_type(self) -> AnyValue:
        """The expression type"""
        pass  # pragma: no cover

    @abstractmethod
    def julia_expression_options(self) -> AnyValue:
        """The expression options"""
        pass  # pragma: no cover

    @abstractmethod
    def create_exports(
        self,
        model: PySRRegressor,
        equations: pd.DataFrame,
        search_output,
    ) -> pd.DataFrame:
        """Create additional columns in the equations dataframe."""
        pass  # pragma: no cover

    @property
    def evaluates_in_julia(self) -> bool:
        return False

    @property
    def supports_sympy(self) -> bool:
        return False

    @property
    def supports_torch(self) -> bool:
        return False

    @property
    def supports_jax(self) -> bool:
        return False

    @property
    def supports_latex(self) -> bool:
        return False


class ExpressionSpec(AbstractExpressionSpec):
    """The default expression specification, with no special behavior."""

    def julia_expression_type(self):
        return SymbolicRegression.Expression

    def julia_expression_options(self):
        return jl.NamedTuple()

    def create_exports(
        self,
        model: PySRRegressor,
        equations: pd.DataFrame,
        search_output,
    ):
        return add_export_formats(
            equations,
            feature_names_in=model.feature_names_in_,
            selection_mask=model.selection_mask_,
            extra_sympy_mappings=model.extra_sympy_mappings,
            extra_torch_mappings=model.extra_torch_mappings,
            output_jax_format=model.output_jax_format,
            extra_jax_mappings=model.extra_jax_mappings,
            output_torch_format=model.output_torch_format,
        )

    @property
    def supports_sympy(self):
        return True

    @property
    def supports_torch(self):
        return True

    @property
    def supports_jax(self):
        return True

    @property
    def supports_latex(self):
        return True


class TemplateExpressionSpec(AbstractExpressionSpec):
    """Spec for templated expressions.

    This class allows you to specify how multiple sub-expressions should be combined
    in a structured way, with constraints on which variables each sub-expression can use.
    Pass this to PySRRegressor with the `expression_spec` argument when you are using
    the `TemplateExpression` expression type.

    Parameters
    ----------
    function_symbols : list[str]
        List of symbols representing the inner expressions (e.g., ["f", "g"]).
        These will be used as keys in the template structure.
    combine : str
        Julia function string that defines how the sub-expressions are combined.
        Takes a NamedTuple of expressions and a tuple of data vectors.
        For example: "((; f, g), (x1, x2, x3)) -> f(x1, x2) + g(x3)^2"
        would constrain f to use x1,x2 and g to use x3.
    num_features : dict[str, int]
        Dictionary mapping function symbols to the number of features each can use.
        For example: {"f": 2, "g": 1} means f takes 2 inputs and g takes 1.
        If not provided, will be inferred from the combine function.

    Examples
    --------
    ```python
    # Create template that combines f(x1, x2) and g(x3):
    expression_spec = TemplateExpressionSpec(
        function_symbols=["f", "g"],
        combine="((; f, g), (x1, x2, x3)) -> sin(f(x1, x2)) + g(x3)^2",
    )

    # Use in PySRRegressor:
    model = PySRRegressor(
        expression_spec=expression_spec
    )
    ```
    """

    def __init__(
        self,
        function_symbols: list[str],
        combine: str,
        num_features: dict[str, int] | None = None,
    ):
        self.function_symbols = function_symbols
        self.combine = combine
        self.num_features = num_features

    def julia_expression_type(self):
        return SymbolicRegression.TemplateExpression

    def julia_expression_options(self):
        f_combine = jl.seval(self.combine)
        creator = jl.seval(
            """
        function _pysr_create_template_structure(
            @nospecialize(function_symbols::AbstractVector),
            @nospecialize(combine::Function),
            @nospecialize(num_features::Union{Nothing,AbstractDict})
        )
            tuple_symbol = (map(Symbol, function_symbols)..., )
            num_features = if num_features === nothing
                nothing
            else
                (; num_features...)
            end
            structure = SymbolicRegression.TemplateStructure{tuple_symbol}(combine, num_features)
            return (; structure)
        end
        """
        )
        return creator(self.function_symbols, f_combine, self.num_features)

    @property
    def evaluates_in_julia(self):
        return True

    def create_exports(
        self,
        model: PySRRegressor,
        equations: pd.DataFrame,
        search_output,
    ) -> pd.DataFrame:
        # We try to load the raw julia state from a saved binary stream
        # if not provided.
        search_output = search_output or model.julia_state_
        return _search_output_to_callable_expressions(equations, search_output)


class ParametricExpressionSpec(AbstractExpressionSpec):
    """Spec for parametric expressions that vary by category.

    This class allows you to specify expressions with parameters that vary across different
    categories in your dataset. The expression structure remains the same, but parameters
    are optimized separately for each category.

    Parameters
    ----------
    max_parameters : int
        Maximum number of parameters that can appear in the expression. Each parameter
        will take on different values for each category in the data.

    Examples
    --------
    For example, if we want to allow for a model with up to 2 parameters (each category
    can have a different value for these parameters), we can use:

    ```python
    model = PySRRegressor(
        expression_spec=ParametricExpressionSpec(max_parameters=2),
        binary_operators=["+", "*"],
        unary_operators=["sin"]
    )
    model.fit(X, y, category=category)
    ```
    """

    def __init__(self, max_parameters: int):
        self.max_parameters = max_parameters

    def julia_expression_type(self):
        return SymbolicRegression.ParametricExpression

    def julia_expression_options(self):
        return jl.seval("NamedTuple{(:max_parameters,)}")((self.max_parameters,))

    @property
    def evaluates_in_julia(self):
        return True

    def create_exports(
        self,
        model: PySRRegressor,
        equations: pd.DataFrame,
        search_output,
    ):
        search_output = search_output or model.julia_state_
        return _search_output_to_callable_expressions(equations, search_output)


class CallableJuliaExpression:
    def __init__(self, expression):
        self.expression = expression

    def __call__(self, X: np.ndarray, *args):
        raw_output = self.expression(jl_array(X.T), *args)
        return np.array(raw_output).T


def _search_output_to_callable_expressions(
    equations: pd.DataFrame, search_output
) -> pd.DataFrame:
    equations = copy.deepcopy(equations)
    (_, out_hof) = search_output
    expressions = []
    callables = []

    for _, row in equations.iterrows():
        curComplexity = row["complexity"]
        expression = out_hof.members[curComplexity - 1].tree
        expressions.append(expression)
        callables.append(CallableJuliaExpression(expression))

    df = pd.DataFrame(
        {"julia_expression": expressions, "lambda_format": callables},
        index=equations.index,
    )
    return df



import json
import sys
from pathlib import Path

import tomlkit

new_backend_version = sys.argv[1]

assert not new_backend_version.startswith("v"), "Version should not start with 'v'"

pyproject_toml = Path(__file__).parent / ".." / ".." / "pyproject.toml"
juliapkg_json = Path(__file__).parent / ".." / ".." / "pysr" / "juliapkg.json"

with open(pyproject_toml) as toml_file:
    pyproject_data = tomlkit.parse(toml_file.read())

with open(juliapkg_json) as f:
    juliapkg_data = json.load(f)

major, minor, patch, *dev = pyproject_data["project"]["version"].split(".")
pyproject_data["project"]["version"] = f"{major}.{minor}.{int(patch)+1}"

juliapkg_data["packages"]["SymbolicRegression"]["version"] = f"={new_backend_version}"

with open(pyproject_toml, "w") as toml_file:
    toml_file.write(tomlkit.dumps(pyproject_data))

with open(juliapkg_json, "w") as f:
    json.dump(juliapkg_data, f, indent=4)
    # Ensure ends with newline
    f.write("\n")



import fnmatch
import sys
import unittest
import warnings

import click

from ..test import (
    get_runtests_cli,
    runtests,
    runtests_dev,
    runtests_jax,
    runtests_startup,
    runtests_torch,
)


@click.group("pysr")
@click.pass_context
def pysr(context):
    ctx = context


@pysr.command("install", help="DEPRECATED (dependencies are now installed at import).")
@click.option(
    "-p",
    "julia_project",
    "--project",
    default=None,
    type=str,
)
@click.option("-q", "--quiet", is_flag=True, default=False, help="Disable logging.")
@click.option(
    "--precompile",
    "precompile",
    flag_value=True,
    default=None,
)
@click.option(
    "--no-precompile",
    "precompile",
    flag_value=False,
    default=None,
)
def _install(julia_project, quiet, precompile):
    warnings.warn(
        "This command is deprecated. Julia dependencies are now installed at first import."
    )


TEST_OPTIONS = {"main", "jax", "torch", "cli", "dev", "startup"}


@pysr.command("test")
@click.argument("tests", nargs=1)
@click.option(
    "-k",
    "expressions",
    multiple=True,
    type=str,
    help="Filter expressions to select specific tests.",
)
def _tests(tests, expressions):
    """Run parts of the PySR test suite.

    Choose from main, jax, torch, cli, dev, and startup. You can give multiple tests, separated by commas.
    """
    test_cases = []
    for test in tests.split(","):
        if test == "main":
            test_cases.extend(runtests(just_tests=True))
        elif test == "jax":
            test_cases.extend(runtests_jax(just_tests=True))
        elif test == "torch":
            test_cases.extend(runtests_torch(just_tests=True))
        elif test == "cli":
            runtests_cli = get_runtests_cli()
            test_cases.extend(runtests_cli(just_tests=True))
        elif test == "dev":
            test_cases.extend(runtests_dev(just_tests=True))
        elif test == "startup":
            test_cases.extend(runtests_startup(just_tests=True))
        else:
            warnings.warn(f"Invalid test {test}. Skipping.")

    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    for test_case in test_cases:
        loaded_tests = loader.loadTestsFromTestCase(test_case)
        for test in loaded_tests:
            if len(expressions) == 0 or any(
                fnmatch.fnmatch(test.id(), "*" + expression + "*")
                for expression in expressions
            ):
                suite.addTest(test)

    runner = unittest.TextTestRunner()
    results = runner.run(suite)

    if not results.wasSuccessful():
        sys.exit(1)



# Example call:
## python3 generate_dev_juliapkg.py /pysr/pysr/juliapkg.json /srjl
import json
import sys

juliapkg_json = sys.argv[1]
path_to_srjl = sys.argv[2]

with open(juliapkg_json, "r") as f:
    juliapkg = json.load(f)

juliapkg["packages"]["SymbolicRegression"] = {
    "uuid": juliapkg["packages"]["SymbolicRegression"]["uuid"],
    "path": path_to_srjl,
    "dev": True,
}

with open(juliapkg_json, "w") as f:
    json.dump(juliapkg, f, indent=4)



"""Functions for doing feature selection during preprocessing."""

from typing import cast

import numpy as np
from numpy import ndarray
from numpy.typing import NDArray

from .utils import ArrayLike


def run_feature_selection(
    X: ndarray,
    y: ndarray,
    select_k_features: int,
    random_state: np.random.RandomState | None = None,
) -> NDArray[np.bool_]:
    """
    Find most important features.

    Uses a gradient boosting tree regressor as a proxy for finding
    the k most important features in X, returning indices for those
    features as output.
    """
    from sklearn.ensemble import RandomForestRegressor
    from sklearn.feature_selection import SelectFromModel

    clf = RandomForestRegressor(
        n_estimators=100, max_depth=3, random_state=random_state
    )
    clf.fit(X, y)
    selector = SelectFromModel(
        clf, threshold=-np.inf, max_features=select_k_features, prefit=True
    )
    return cast(NDArray[np.bool_], selector.get_support(indices=False))


# Function has not been removed only due to usage in module tests
def _handle_feature_selection(
    X: ndarray,
    select_k_features: int | None,
    y: ndarray,
    variable_names: ArrayLike[str],
):
    if select_k_features is not None:
        selection = run_feature_selection(X, y, select_k_features)
        print(f"Using features {[variable_names[i] for i in selection]}")
        X = X[:, selection]
    else:
        selection = None

    return X, selection



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from .julia_helpers import jl_array, jl_dict
from .julia_import import AnyValue, jl


class AbstractLoggerSpec(ABC):
    """Abstract base class for logger specifications."""

    @abstractmethod
    def create_logger(self) -> AnyValue:
        """Create a logger instance."""
        pass  # pragma: no cover

    @abstractmethod
    def write_hparams(self, logger: AnyValue, hparams: dict[str, Any]) -> None:
        """Write hyperparameters to the logger."""
        pass  # pragma: no cover

    @abstractmethod
    def close(self, logger: AnyValue) -> None:
        """Close the logger instance."""
        pass  # pragma: no cover


@dataclass
class TensorBoardLoggerSpec(AbstractLoggerSpec):
    """Specification for TensorBoard logger.

    Parameters
    ----------
    log_dir : str
        Directory where TensorBoard logs will be saved. If `overwrite` is `False`,
        new logs will be saved to `{log_dir}_1`, and so on. Default is `"logs/run"`.
    log_interval : int, optional
        Interval (in steps) at which logs are written. Default is 10.
    overwrite : bool, optional
        Whether to overwrite existing logs in the directory. Default is False.
    """

    log_dir: str = "logs/run"
    log_interval: int = 1
    overwrite: bool = False

    def create_logger(self) -> AnyValue:
        # We assume that TensorBoardLogger is already imported via `julia_extensions.py`
        make_logger = jl.seval(
            """
            function make_logger(log_dir::AbstractString, overwrite::Bool, log_interval::Int)
                base_logger = TensorBoardLogger.TBLogger(
                    log_dir,
                    (overwrite ? (TensorBoardLogger.tb_overwrite,) : ())...
                )
                return SRLogger(; logger=base_logger, log_interval)
            end
        """
        )
        log_dir = str(self.log_dir)
        return make_logger(log_dir, self.overwrite, self.log_interval)

    def write_hparams(self, logger: AnyValue, hparams: dict[str, Any]) -> None:
        base_logger = jl.SymbolicRegression.get_logger(logger)
        writer = jl.seval("TensorBoardLogger.write_hparams!")
        jl_clean_hparams = jl_dict(
            {
                k: (v if isinstance(v, (bool, int, float)) else str(v))
                for k, v in hparams.items()
            }
        )
        writer(
            base_logger,
            jl_clean_hparams,
            jl_array(
                [
                    "search/data/summaries/pareto_volume",
                    "search/data/summaries/min_loss",
                ],
            ),
        )

    def close(self, logger: AnyValue) -> None:
        base_logger = jl.SymbolicRegression.get_logger(logger)
        jl.close(base_logger)



import os

from setuptools import setup

if os.path.exists(".git"):
    kwargs = {
        "use_scm_version": {
            "write_to": "pysr/version.py",
        },
        "setup_requires": ["setuptools", "setuptools_scm"],
    }
else:
    # Read from pyproject.toml directly
    import re

    with open(os.path.join(os.path.dirname(__file__), "pyproject.toml")) as f:
        data = f.read()
        # Find the version
        version = re.search(r'version = "(.*)"', data).group(1)

    # Write the version to version.py
    with open(os.path.join(os.path.dirname(__file__), "pysr", "version.py"), "w") as f:
        f.write(f'__version__ = "{version}"')

    kwargs = {
        "use_scm_version": False,
        "version": version,
    }


# Build options are managed in pyproject.toml
setup(**kwargs)



"""Various functions to deprecate features."""

import warnings

from .julia_import import jl


def install(*args, **kwargs):
    del args, kwargs
    warnings.warn(
        "The `install` function has been removed. "
        "PySR now uses the `juliacall` package to install its dependencies automatically at import time. ",
        FutureWarning,
    )


def init_julia(*args, **kwargs):
    del args, kwargs
    warnings.warn(
        "The `init_julia` function has been removed. "
        "Julia is now initialized automatically at import time.",
        FutureWarning,
    )
    return jl


def pysr(X, y, weights=None, **kwargs):  # pragma: no cover
    from .sr import PySRRegressor

    warnings.warn(
        "Calling `pysr` is deprecated. "
        "Please use `model = PySRRegressor(**params); "
        "model.fit(X, y)` going forward.",
        FutureWarning,
    )
    model = PySRRegressor(**kwargs)
    model.fit(X, y, weights=weights)
    return model.equations_


def best(*args, **kwargs):  # pragma: no cover
    raise NotImplementedError(
        "`best` has been deprecated. "
        "Please use the `PySRRegressor` interface. "
        "After fitting, you can return `.sympy()` "
        "to get the sympy representation "
        "of the best equation."
    )


def best_row(*args, **kwargs):  # pragma: no cover
    raise NotImplementedError(
        "`best_row` has been deprecated. "
        "Please use the `PySRRegressor` interface. "
        "After fitting, you can run `print(model)` to view the best equation, "
        "or "
        "`model.get_best()` to return the best equation's "
        "row in `model.equations_`."
    )


def best_tex(*args, **kwargs):  # pragma: no cover
    raise NotImplementedError(
        "`best_tex` has been deprecated. "
        "Please use the `PySRRegressor` interface. "
        "After fitting, you can return `.latex()` to "
        "get the sympy representation "
        "of the best equation."
    )


def best_callable(*args, **kwargs):  # pragma: no cover
    raise NotImplementedError(
        "`best_callable` has been deprecated. Please use the `PySRRegressor` "
        "interface. After fitting, you can use "
        "`.predict(X)` to use the best callable."
    )


DEPRECATED_KWARGS = {
    "fractionReplaced": "fraction_replaced",
    "fractionReplacedHof": "fraction_replaced_hof",
    "npop": "population_size",
    "hofMigration": "hof_migration",
    "shouldOptimizeConstants": "should_optimize_constants",
    "weightAddNode": "weight_add_node",
    "weightDeleteNode": "weight_delete_node",
    "weightDoNothing": "weight_do_nothing",
    "weightInsertNode": "weight_insert_node",
    "weightMutateConstant": "weight_mutate_constant",
    "weightMutateOperator": "weight_mutate_operator",
    "weightSwapOperands": "weight_swap_operands",
    "weightRandomize": "weight_randomize",
    "weightSimplify": "weight_simplify",
    "crossoverProbability": "crossover_probability",
    "perturbationFactor": "perturbation_factor",
    "batchSize": "batch_size",
    "warmupMaxsizeBy": "warmup_maxsize_by",
    "useFrequency": "use_frequency",
    "useFrequencyInTournament": "use_frequency_in_tournament",
    "ncyclesperiteration": "ncycles_per_iteration",
    "loss": "elementwise_loss",
    "full_objective": "loss_function",
}
